/*calculate probabilities for gene clusters 
  1.Ignore gene orientations
  2.The larger the size of gene clusters, the smaller the E-value
  3.For query only

  how to calculate:
  1. calculate p(n, n', k, k') -- probability of k'hits within a window of size k
     n -- number of genes in a genome; 
	 n'-- number of genes associated with a query gene (#hits)
	 k -- cluster length, or window size
	 k'-- number of genes in a cluster
  2. calculate e(n, n', k, k') -- the expected number of such clusters. 
**/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>


#define FPFE1(x)    fprintf(stdout, "%s", x), fflush(NULL), exit(0)
#define FPFE2(x, y) fprintf(stdout, x, y), fflush(NULL), exit(0)

int OUTPUT_HEADER = 0;

int isCircularChromosome = 0;

//calculate all e-values
void comput_evalues(int n, int np, int l, double *** EVALUES, int chrform);


//----------------calculate probabilities for break points---------------->>>>>>
typedef struct bigdouble {
	double d;
	double e;
} bigdouble;

bigdouble bigdouble_mul(bigdouble a, bigdouble b){
	bigdouble c;
	double x;
	int y;

	if( a.d == 0.0 || b.d == 0.0 ){
		c.d = 0.0;
		c.e = 0.0;
	} else {
		c.e = a.e + b.e;
		x = a.d * b.d;

		y = 0;
		if( fabs(x) >= 10.0 ){
			while(fabs(x) >= 10.0){
				x /= 10.0;
				y++;
			}
		} else if ( fabs(x) < 1.0 ) {
			if( fabs(x) == 0.0 )
				fprintf(stdout, "\nHerem x =0 "), fflush(NULL), exit(0);

			while( fabs(x) < 1.0){
				x *= 10.0;
				y--;
			}
		}

		c.e += y;
		c.d  = x;
	}

	return c;	
}

bigdouble bigdouble_mul3(bigdouble a, bigdouble b, bigdouble c){
	return bigdouble_mul(bigdouble_mul(a, b), c);
}

bigdouble bigdouble_div(bigdouble a, bigdouble b){
	bigdouble c;
	double x;
	int y;

	if ( a.d == 0 ){
		return a;
	} else if (b.d == 0){
		fprintf(stdout, "Divided by 0!"), fflush(NULL), exit(0);
	} else {
		c.e = a.e - b.e;
		x = a.d / b.d;

		y = 0;
		if( fabs(x) >= 10.0 ){
			while(fabs(x) >= 10.0){
				x /= 10.0;
				y++;
			}
		} else if ( fabs(x) < 1.0 ) {
			if( fabs(x) == 0.0 )
				fprintf(stdout, "\nHerem x =0 "), fflush(NULL), exit(0);
			
			while( fabs(x) < 1.0){
				x *= 10.0;
				y--;
			}
		}

		c.e += y;
		c.d  = x;
	}

	return c;	
}

bigdouble bigdouble_add(bigdouble a, bigdouble b){
	bigdouble c;
	double x;
	int y;

	if( a.d == 0 )
		return b;

	if( b.d == 0 )
		return a;

	if( a.e == b.e ){
		c.e = a.e;
		x = a.d + b.d;
	} else if ( a.e > b.e ) {
		c.e = a.e;
		x = a.d + b.d * pow(10.0, b.e - a.e);
	} else {
		c.e = b.e;
		x = b.d + a.d * pow(10.0, a.e - b.e);
	}

	y = 0;
	if( fabs(x) >= 10.0 ){
		while(fabs(x) >= 10.0){
			x /= 10.0;
			y++;
		}
	} else if ( fabs(x) < 1.0 ) {
		if( fabs(x) == 0.0 )
			fprintf(stdout, "\nHerem x =0 "), fflush(NULL), exit(0);

		while( fabs(x) < 1.0){
			x *= 10.0;
			y--;
		}
	}

	c.e += y;
	c.d  = x;

	return c;	
}


bigdouble bigdouble_sub(bigdouble a, bigdouble b){
	bigdouble c;
	double x;
	int y;

	if( a.d == 0 ){
		b.d *= (-1);
		return b;
	}

	if( b.d == 0 )
		return a;

	if( a.e == b.e ){
		c.e = a.e;
		x = a.d - b.d;
	} else if ( a.e > b.e ) {
		c.e = a.e;
		x = a.d - b.d * pow(10.0, b.e - a.e);
	} else {
		c.e = b.e;
		x = a.d * pow(10.0, a.e - b.e) - b.d;
	}

	y = 0;
	if( fabs(x) >= 10.0 ){
		while(fabs(x) >= 10.0){
			x /= 10.0;
			y++;
		}
	} else if ( fabs(x) < 1.0 ) {
		if( fabs(x) == 0.0 )
			fprintf(stdout, "\nHerem x =0 "), fflush(NULL), exit(0);

		while( fabs(x) < 1.0){
			x *= 10.0;
			y--;
		}
	}

	c.e += y;
	c.d  = x;

	return c;	
}


//calculate combination
void comb(bigdouble ** c, int l){
	bigdouble * b; 
	double x;
	int y, k;

	if( !(b= calloc(l+1, sizeof(bigdouble))) )
		FPFE1("out of memory for b");
	
	b[0].d = 1.0;
	b[0].e = 0;
	for(k=1; k<=l/2; k++){
		x = (double)(l-k+1)/k;

		if( x == 0.0 )
			fprintf(stdout, "\nl=%d k=%d", l, k), fflush(NULL), exit(0);

		y = 0;
		if( x >= 10.0 ){
			while(x >= 10.0){
				x /= 10.0;
				y++;
			}
		} else if ( x < 1.0 ) {
			while( x < 1.0){
				x *= 10.0;
				y--;
			}
		}	
		b[k].e = b[k-1].e + y;
		b[k].d = b[k-1].d * x;

		if( b[k].d >= 10.0 ){
			b[k].d /= 10.0;
			b[k].e += 1;
		}
	}

	for(k=l/2+1; k<=l; k++){
		b[k] = b[l-k];
	}

	*c = b;

	return;
}

//calculate 2-d combination -- k, k'
void comb2d(bigdouble *** c, int l){
	bigdouble **d;

	if( !(d = (bigdouble **)calloc(l+1, sizeof(bigdouble *))) )
		FPFE1("out of memory for d");

	while(l >= 0){
		comb(d+l, l);
		l--;
	}

	*c = d;
	return;
}

/*
//calculate p(n, n', k, k') 
void comput_pkkprime(bigdouble **com, int l, int n, int np, bigdouble *** pmk){
	bigdouble ** mk;
	int i, j,nd;
	bigdouble x;

	if( !(*pmk = mk = (bigdouble **)calloc(l+1, sizeof(bigdouble *))) )
		FPFE1("out of memory for mk");
	for( i=0; i<=l; i++){
		if( !(mk[i] = (bigdouble *)calloc(i+1, sizeof(bigdouble))) )
			FPFE1("out of memory for mk[i]");
	}

	nd = n - np;

	for(i=0; i<=l; ++i){ //k
		x.d = n-i+1;
		x.e = 0.0;

		mk[i][i] = bigdouble_div(com[np][i], com[n][i]);
		for(j=i-1; j>=0; --j){ //k'
			mk[i][j] = bigdouble_add(mk[i][j+1], bigdouble_div(bigdouble_mul(com[np][j], com[nd][i-j]), com[n][i]));
		}

		for(j=i; j>=0; --j){ //k'
			mk[i][j] = bigdouble_mul(mk[i][j], x);
		}
	}

	return;
}


//get evalues in simple format
void cal_evalues(bigdouble **d, int l, double *** evals){
	double ** vals;
	int i, j, n;

	n = (int)log10(DBL_MAX);

	if( !(*evals = vals = (double **)calloc(l+1, sizeof(double *))) )
		FPFE1("out of memory for vals");
	for(i=0; i<=l; i++)
		if( !(vals[i] = (double *)calloc(i+1, sizeof(double))) )
			FPFE1("out of memory for vals[i]");

	for(i=0; i<=l; i++){
		for(j=0; j<=i; j++){	
			if( d[i][j].e >= n-2 ){
				vals[i][j] = DBL_MAX; 
			} else if ( d[i][j].e <= 2-n ){
				vals[i][j] = 0.0; 
			} else {
				vals[i][j] = pow(10.0, d[i][j].e) * d[i][j].d;			
			}
		}
	}

	return;
}
*/

//calculate p(n, n', k, k') 
void comput_pkkprime(bigdouble **com, int l, int n, int np, bigdouble *** pmk){
	bigdouble ** mk;
	int i, j,nd, mm;
	bigdouble x;

	if( !(*pmk = mk = (bigdouble **)calloc(l+1, sizeof(bigdouble *))) )
		FPFE1("out of memory for mk");
	for( i=0; i<=l; i++){
		mm = (i>np) ? (np) : (i);

		//if( !(mk[i] = (bigdouble *)calloc(i+1, sizeof(bigdouble))) )
		if( !(mk[i] = (bigdouble *)calloc(mm+1, sizeof(bigdouble))) )
			FPFE1("out of memory for mk[i]");
	}

	nd = n - np;

	for(i=0; i<=l; ++i){ //k
		if(isCircularChromosome == 0)
			x.d = n;
		else{
			x.d = n-i+1;
		}

		x.e = 0.0;

		mm = (i>np) ? (np) : (i);
		mk[i][mm] = bigdouble_div(bigdouble_mul(com[np][mm], com[nd][i-mm]), com[n][i]);

		for(j=mm-1; j>=0; --j){ //k'
			mk[i][j] = bigdouble_add(mk[i][j+1], bigdouble_div(bigdouble_mul(com[np][j], com[nd][i-j]), com[n][i]));
		}

		for(j=mm; j>=0; --j){ //k'
			mk[i][j] = bigdouble_mul(mk[i][j], x);
		}
	}

	return;
}


//get evalues in simple format
void cal_evalues(bigdouble **d, int l, int np, double *** evals){
	double ** vals;
	int i, j, mm;

	int n;
	n = (int)log10(DBL_MAX);	

	if( !(*evals = vals = (double **)calloc(l+1, sizeof(double *))) )
		FPFE1("out of memory for vals");
	for(i=0; i<=l; i++){
		mm = (i>np) ? (np) : (i);

		if( !(vals[i] = (double *)calloc(mm+1, sizeof(double))) )
			FPFE1("out of memory for vals[i]");
	}

	for(i=0; i<=l; i++){
		mm = (i>np) ? (np) : (i);

		for(j=0; j<=mm; j++){		
			if( d[i][j].e >= n-2 ){
				vals[i][j] = DBL_MAX; 
			} else if ( d[i][j].e <= 2-n ){
				vals[i][j] = 0.0; 
			} else {
				vals[i][j] = pow(10.0, d[i][j].e) * d[i][j].d;			
			}
		}
	}

	return;
}


//calculate all e-values
//n -- number of genes on a genome
//np -- number of hits
//l -- max window size
void comput_evalues(int n, int np, int l, double *** EVALUES, int chrform){
	bigdouble ** comnn;
	bigdouble ** pkkprime;  //p(n, n', k, k')

	isCircularChromosome = chrform;
	
	//calculate 2-d combination -- n -- total# of genes on a genome
	comb2d(&comnn, n);

	//calculate p(n, n', k, k') 
	comput_pkkprime(comnn, l, n, np, &pkkprime);

	//get evalues in simple format
	cal_evalues(pkkprime, l, np, EVALUES);

/*
{
	FILE *outf;
	int i,j;

	if( !(outf = fopen("testeval", "w")))
		FPFE1("Can't open file for writing!");
	for(i=0; i<=l; i++){
		fprintf(outf, "\nl=%5d", i);
		for(j=0; j<=i; j++){			
			fprintf(outf, " %12e", (*EVALUES)[i][j]);			
		}
	}
	fclose(outf);
}
	fflush(NULL);
//*/


	return;
}


//----------------calculate probabilities for break points----------------<<<<<<

/*
int main(){
	int l, n, np;
  

	n=2000;
	np=800;
	l=200;
	
	//calculate all e-values
	//n -- number of orthologs (no paralog)
	//l -- max window size
	comput_evalues(n, np, l, &EVALUES);
}
*/


