#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <float.h>

//#define	DEBUG	1

int LONGEST_PROTEIN_LEN = 20000;
double E_VALUE_CUTOFF = 1e-5;
int E_LENGTH_CUTOFF = 1;
double GENE_CLUSTER_E_VALUE_CUTOFF = 1e-5;

//protein structure
typedef struct T_PROTEIN {
	char * name;
	char * seq;
} T_PROTEIN;


//position structure (store information for a position on the chromosome
typedef struct T_POSITION {
	char * name;
	int    pos;	
	double * score;
	struct T_POSITION * next;
} T_POSITION;

//result cluster
typedef struct T_GCLUSTER {
	double e;        //evalue
	int n;           //size of the cluster
	T_POSITION ** ps; //pos
	struct T_GCLUSTER * next;
} T_GCLUSTER;

//check if a line is blank
int isblank(char * s){
	while( *s != '\0' && *s != EOF ){
		if( isalnum(*s) )
			return 0;
		s++;
	}
	return 1;
}

int file_exists(const char * filename){
	FILE * file;
    if ( file = fopen(filename, "r"))
    {
        fclose(file);
        return 1;
    }
    return 0;
}


//read in only names of protein sequences ( in FASTA format )  
void sequence_reader_fasta_name_only(char *f, T_PROTEIN **ps, int *n_ps ) {
	FILE * inf;
	char * buff, *p; //[LONGEST_SEQ_LEN];

	T_PROTEIN * prs;
	int nprs;


	if( !(inf = fopen(f, "r")) )
		fprintf(stdout, "File %s can not be opened for reading!\n", f), fflush(NULL), exit(0);

	if( !(buff = malloc(LONGEST_PROTEIN_LEN * sizeof(char))) )
		fprintf(stdout, "\n out of memory for buff!\n"), fflush(NULL), exit(0);

	nprs = 0;
	while( fgets(buff, LONGEST_PROTEIN_LEN, inf) ){
		if( strstr(buff, ">") )
			++nprs;
	}

	*n_ps = nprs;
	if( !(*ps = prs = calloc(nprs, sizeof(T_PROTEIN))) )
		fprintf(stdout, "\n out of memory for prs!\n"), fflush(NULL), exit(0);

	rewind(inf);

	nprs = -1;
	while( fgets(buff, LONGEST_PROTEIN_LEN, inf) ){
		if( strstr(buff, ">") ){
			++nprs;

			if( buff[strlen(buff)-1] == '\n' )
				buff[strlen(buff)-1] = '\0';

			p = strstr(buff, ">")+1;
            
			if( !(prs[nprs].name = malloc((strlen(p)+1)*sizeof(char))) )
				fprintf(stdout, "\n out of memory for prs[nprs]!\n"), fflush(NULL), exit(0);

			strcpy(prs[nprs].name, p);			
		} 
	}
		
	fclose(inf);

	if( buff )
		free(buff);

	return;
}



//read in protein sequence in FASTA format
void sequence_reader_fasta(char *f, T_PROTEIN **ps, int *n_ps ) {
	FILE * inf;
	char * buff; //[LONGEST_SEQ_LEN];
	char * tmpstr;
	int    n_tmpstr;

	char *p;

	T_PROTEIN * prs;
	int nprs;


	if( !(inf = fopen(f, "r")) )
		fprintf(stdout, "File %s can not be opened for reading!\n", f), fflush(NULL), exit(0);

	if( !(buff = malloc(LONGEST_PROTEIN_LEN * sizeof(char))) )
		fprintf(stdout, "\n out of memory for buff!\n"), fflush(NULL), exit(0);

	if( !(tmpstr = malloc(LONGEST_PROTEIN_LEN * sizeof(char))) )
		fprintf(stdout, "\n out of memory for tmpstr!\n"), fflush(NULL), exit(0);

	nprs = 0;
	while( fgets(buff, LONGEST_PROTEIN_LEN, inf) ){
		if( strstr(buff, ">") )
			++nprs;
	}

	*n_ps = nprs;
	if( !(*ps = prs = calloc(nprs, sizeof(T_PROTEIN))) )
		fprintf(stdout, "\n out of memory for prs!\n"), fflush(NULL), exit(0);

	rewind(inf);

	nprs = -1;
	n_tmpstr = 0;
	while( fgets(buff, LONGEST_PROTEIN_LEN, inf) ){
		if( strstr(buff, ">") ){
			if( nprs >= 0 ){
				tmpstr[n_tmpstr] = '\0';
				if( !(prs[nprs].seq = malloc((strlen(tmpstr)+1)*sizeof(char))) )
						fprintf(stdout, "\n out of memory for prs[nprs].seq!\n"), fflush(NULL), exit(0);

				strcpy(prs[nprs].seq, tmpstr);			
			}

			++nprs;
			n_tmpstr = 0;

			if( buff[strlen(buff)-1] == '\n' )
				buff[strlen(buff)-1] = '\0';

			p = strstr(buff, ">")+1;
           
			if( !(prs[nprs].name = malloc((strlen(p)+1)*sizeof(char))) )
				fprintf(stdout, "\n out of memory for prs[nprs]!\n"), fflush(NULL), exit(0);

			strcpy(prs[nprs].name, p);			
		} else {
			p = buff;
			while( *p != '\0' && *p != EOF  ){
				if( isalpha(*p) ){
					tmpstr[n_tmpstr] = toupper(*p);
					n_tmpstr ++;					
				}

				p++;
			}
		}
	}

	//copy the last sequence
	if( nprs >= 0 ){
		tmpstr[n_tmpstr] = '\0';
		if( !(prs[nprs].seq = malloc((strlen(tmpstr)+1)*sizeof(char))) )
				fprintf(stdout, "\n out of memory for prs[nprs].seq!\n"), fflush(NULL), exit(0);

		strcpy(prs[nprs].seq, tmpstr);			
	}
		
	fclose(inf);

	if( buff )
		free(buff);
	if( tmpstr )
		free(tmpstr);

	return;
}


//convert a .faa file to a new .faa file with different names
void faa_file_converter(char *f, char *newf ) {
	FILE * inf, *outf;
	char * buff; //[LONGEST_SEQ_LEN];
	int nprs;

	if( !(inf = fopen(f, "r")) )
		fprintf(stdout, "File %s can not be opened for reading!\n", f), fflush(NULL), exit(0);

	if( !(outf = fopen(newf, "w")) )
		fprintf(stdout, "File %s can not be opened for writing!\n", newf), fflush(NULL), exit(0);

	if( !(buff = malloc(LONGEST_PROTEIN_LEN * sizeof(char))) )
		fprintf(stdout, "\n out of memory for buff!\n"), fflush(NULL), exit(0);

	nprs = 0;
	while( fgets(buff, LONGEST_PROTEIN_LEN, inf) ){
		if( strstr(buff, ">") ){
			fprintf(outf, ">%1d\n", nprs);
			++nprs;
		} else {
			fprintf(outf, "%s", buff);			
		}
	}
	
	fclose(inf);
	fclose(outf);

	if( buff )
		free(buff);

	return;
}


//process a chromosome in a .faa file and get a pointer to T_POSITION containing position information etc. for next step
void process_chromosome(char *faa, T_PROTEIN * gcps, int n_gcps, T_POSITION ** pos, int *ngenes){
	FILE * inf, *outf;

#ifdef DEBUG
	FILE * outlog;
#endif

	char * buff; //[LONGEST_SEQ_LEN];
	int    i,j, n, k;

	T_PROTEIN * prs;
	int nprs; //number of proteins on the chromosome
	char tmpfaafile[50], tmpseqfile[50], tmpoutfile[50], tmpcmd[300];
	char tmpstr[200];

	T_POSITION * chrpos = NULL, *tail=NULL, *curr, *durr;
	T_POSITION * newpos;

	double * tmpscores=NULL;
	double tmpevalue;

	//read in all protein names 
	//read in only names of protein sequences ( in FASTA format )  
	sequence_reader_fasta_name_only(faa, &prs, &nprs);

	//create a tmp .faa file contain the original proteins
	strcpy(tmpfaafile, "tmpfaafile_xxx_tobedeleted");
	strcpy(tmpseqfile, "tmpseqfile_xxx_tobedeleted");
	strcpy(tmpoutfile, "tmpoutfile_xxx_tobedeleted");
	//convert a .faa file to a new .faa file with different names
	if( ! file_exists("tmpfaafile_xxx_tobedeleted") )
		faa_file_converter(faa, tmpfaafile);

	if( !(file_exists("tmpfaafile_xxx_tobedeleted")) ){		
		fprintf(stderr, "\n\nCannot use formatdb to generate database files for blast!\n\n"), fflush(NULL), exit(0);
	} 

	//call formatdb 
	if( !( file_exists("tmpfaafile_xxx_tobedeleted.phr") 
		&& file_exists("tmpfaafile_xxx_tobedeleted.pin") 
		&& file_exists("tmpfaafile_xxx_tobedeleted.psq")) ){
		system("formatdb -i tmpfaafile_xxx_tobedeleted -p T");
	}

	if( !(file_exists("tmpfaafile_xxx_tobedeleted.phr") && file_exists("tmpfaafile_xxx_tobedeleted.pin") && file_exists("tmpfaafile_xxx_tobedeleted.psq")) ){
		remove("tmpfaafile_xxx_tobedeleted.phr");
		remove("tmpfaafile_xxx_tobedeleted.pin");
		remove("tmpfaafile_xxx_tobedeleted.psq");
		remove("tmpfaafile_xxx_tobedeleted");

		fprintf(stderr, "\n\nCannot use formatdb to generate database files for blast!\n\n"), fflush(NULL), exit(0);
	} 

	if( !(buff = malloc(LONGEST_PROTEIN_LEN * sizeof(char))) )
		fprintf(stdout, "\n out of memory for buff!\n"), fflush(NULL), exit(0);

#ifdef DEBUG
	if( !(outlog = fopen("debuglog", "a")) )
		fprintf(stdout, "File debuglog cannot be opened for writing!\n"), fflush(NULL), exit(0);
#endif

	//call blastall to do the query 
	for(i=0; i<n_gcps; ++i){ //for each query sequence
		if( !(outf = fopen(tmpseqfile, "w")) )
			fprintf(stdout, "File %s can not be opened for writing!\n", tmpseqfile), fflush(NULL), exit(0);
		fprintf(outf, ">%d\n%s\n", i, gcps[i].seq);
		fclose(outf);

		sprintf(tmpcmd, "blastall -p blastp -d tmpfaafile_xxx_tobedeleted -i %s -o %s ", tmpseqfile, tmpoutfile, i);
		system(tmpcmd);

		//process blastall results
		if( !(inf = fopen(tmpoutfile, "r")) )
			fprintf(stdout, "File %s can not be opened for reading!\n", tmpoutfile), fflush(NULL), exit(0);

#ifdef DEBUG
		fprintf(outlog, "\n\n****************************************************\n\n");

		while( fgets(buff, LONGEST_PROTEIN_LEN, inf) ){
			fprintf(outlog, "%s", buff);
		}
		rewind(inf);
#endif

		while( fgets(buff, LONGEST_PROTEIN_LEN, inf) && !strstr(buff, "Sequences producing significant alignments:") ){
		}

		while( fgets(buff, LONGEST_PROTEIN_LEN, inf) && !strstr(buff, ">") ){
			//if( sscanf(buff, " %d %d %le ", &n, &k, &tmpevalue) == 3 ){
			if( sscanf(buff, " %d %d %s ", &n, &k, tmpstr ) == 3 ){

				//if( tmpstr[0] == 'e' ){ //e-173
				//	tmpevalue = 
				//}

				tmpevalue = atof(tmpstr);


#ifdef DEBUG
				fprintf(outlog, "%e\n", tmpevalue);		
#endif


				if( tmpevalue > E_VALUE_CUTOFF )
					break;

				if( chrpos ){
					curr = chrpos;

					while( curr && curr->pos < n+1 ){
						curr = curr->next;
					}
					
					if(  curr && curr->pos == n+1 ){
						curr->score[i] = tmpevalue;
						continue;
					}

				}

				if( !(newpos = malloc(sizeof(T_POSITION))) )
					fprintf(stdout, "\n out of memory for newpos!\n"), fflush(NULL), exit(0);

				//name
				if( !(newpos->name = malloc((strlen(prs[n].name)+1)*sizeof(char))) )
					fprintf(stdout, "\n out of memory for newpos->name!\n"), fflush(NULL), exit(0);

				strcpy(newpos->name, prs[n].name);

				//position
				newpos->pos = n+1;

				//next
				newpos->next = NULL;

				//scores
				if( !(newpos->score = malloc(n_gcps*sizeof(double))) )
					fprintf(stdout, "\n out of memory for newpos->score!\n"), fflush(NULL), exit(0);				
				for(j=0; j<n_gcps; ++j){
					newpos->score[j] = FLT_MAX;						
				}
				newpos->score[i] = tmpevalue;	

				if( ! chrpos ){							
					chrpos = tail = newpos;
				} else {
					if( curr == chrpos ){
						newpos->next = chrpos;
						chrpos = newpos;
					} else if( !curr ){
						tail->next = newpos;
						tail = tail->next;
					} else {
						durr = chrpos;
						while( durr->next != curr )
							durr = durr->next;

						durr->next = newpos;
						newpos->next = curr;
					}					
				}
				
			}
		}

		fclose(inf);
		remove(tmpoutfile);
		remove(tmpseqfile);		
	}

	*ngenes = nprs; 
	*pos = chrpos;

	if( buff )
		free(buff);
	if( prs ){
		for(i=0; i<nprs; ++i){
			if(prs[i].name)
				free(prs[i].name);
		}
		free(prs);
	}

	//remove("tmpfaafile_xxx_tobedeleted.phr");
	//remove("tmpfaafile_xxx_tobedeleted.pin");
	//remove("tmpfaafile_xxx_tobedeleted.psq");
	//remove("tmpfaafile_xxx_tobedeleted");

#ifdef DEBUG
	fclose(outlog);
#endif


	return;
}


//get file names containing input data
void get_file_names(char * fname, char *** filenames, int *n_filenames){
	FILE *inf;
	char buff[4096], *ptr;
	char ** gns;
	int n=0;

	if( !(inf = fopen(fname, "r")) )
		fprintf(stderr, "File %1s can't be opened for reading", fname), fflush(NULL), exit(0);
	while( fgets(buff, 4096, inf) )
		if( !isblank(buff) )
			n++;

	*n_filenames = n;
	if( !(gns = malloc(n * sizeof(char *))) )
		fprintf(stderr, "out of memory for *filenames!\n"), fflush(NULL), exit(0);

	rewind(inf);
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( !isblank(buff) ){
			ptr=buff;
			while( *ptr != '\0' && *ptr != '\n' && *ptr != EOF )
				ptr++;
			if( *ptr == '\n' || *ptr == EOF )
				*ptr = '\0';

			if( !( gns[n] = malloc( (strlen(buff)+1) * sizeof(char))) )
				fprintf(stderr, "out of memory for *filenames[]!\n");
			strcpy(gns[n], buff);
			n++;
		}
	}
	*filenames = gns;
	fclose(inf);
	return;
}


//insert a new cluster into the resulting cluster
void insert_cluster(T_POSITION * p, T_POSITION * p2, double ** evals, int currwinsize, int currn, T_GCLUSTER ** clusters){
	int  j;

	T_POSITION * curr; 
	T_GCLUSTER  * newcluster, *tmp, *turr, *durr; 

	if( !(newcluster = malloc(sizeof(T_GCLUSTER))) )
		fprintf(stderr, "out of memory for newcluster!\n"), fflush(NULL), exit(0);

	if( !(newcluster->ps = malloc(currn * sizeof(T_POSITION *))) )
		fprintf(stderr, "out of memory for newcluster.ps!\n"), fflush(NULL), exit(0);

	newcluster->e = evals[currwinsize][currn];
	newcluster->n = currn;

	j = 0;
	curr = p;					
	while(curr != p2){
		newcluster->ps[j] = curr;

		++j;
		curr = curr->next;
	}
	newcluster->ps[j] = curr;

	newcluster->next = NULL;

	if( *clusters == NULL ){
		*clusters = newcluster;
	} else {
		tmp = *clusters;

		if( tmp->e > newcluster->e ){
			newcluster->next = tmp;
			*clusters = newcluster;
		} else {
			while(tmp->next && tmp->next->e <= newcluster->e){
				tmp = tmp->next;
			}

			newcluster->next = tmp->next;
			tmp->next = newcluster;
		}

		//remove subsets
		tmp = newcluster;

		while( tmp->next ){
			durr = tmp->next;

			if(newcluster->ps[0]->pos <= newcluster->ps[newcluster->n -1]->pos)
			{
				if( (durr->ps[0]->pos <= durr->ps[durr->n -1]->pos) && 
					(newcluster->ps[0]->pos <= durr->ps[0]->pos) && 
					(newcluster->ps[newcluster->n -1]->pos >= durr->ps[durr->n -1]->pos) )
				{
					turr = durr; 
					tmp->next = durr->next;

					if( turr->ps )
						free( turr->ps );
					if( turr )
						free( turr );

					continue;				
				}			
			}
			else
			{
				if((newcluster->ps[0]->pos <= durr->ps[0]->pos && newcluster->ps[newcluster->n -1]->pos >= durr->ps[durr->n -1]->pos) || 
				   (newcluster->ps[0]->pos <= durr->ps[0]->pos && newcluster->ps[0]->pos <= durr->ps[durr->n -1]->pos) || 
				   (durr->ps[0]->pos <= newcluster->ps[newcluster->n -1]->pos && durr->ps[durr->n -1]->pos<= newcluster->ps[newcluster->n -1]->pos))					
				{
					turr = durr; 
					tmp->next = durr->next;

					if( turr->ps )
						free( turr->ps );
					if( turr )
						free( turr );

					continue;	
				}		
			}

			//if( durr->ps[0]->pos >= newcluster->ps[0]->pos && durr->ps[durr->n -1]->pos <= newcluster->ps[newcluster->n -1]->pos){
			//	turr = durr; 
			//	tmp->next = durr->next;

			//	if( turr->ps )
			//		free( turr->ps );
			//	if( turr )
			//		free( turr );

			//	continue;
			//}

			tmp = tmp->next;
		}
	}

	return;
}

//find gene cluster in a chromosome and get a pointer to T_RESULT containing resulting clusters
void find_clusters_inonechr(T_POSITION * pos, double ** evals, int maxwinsize, T_GCLUSTER ** clusters){
	int  i, n;

	T_POSITION * curr, *p, *p2; 

	for(i=1; i<maxwinsize; ++i){ //window size = i+1
		p = pos;
		p2 = pos;

		while( p2 ){
			while(p2 && p2->pos - p->pos < i)
				p2 = p2->next;

			if( !p2 )
				break;

			if( p2->pos - p->pos == i ){ //meet a cluster with window_size=i+1
				curr = p;
				n = 1;

				while(curr != p2){
					++n;
					curr = curr->next;
				}

				if( evals[i+1][n] <= GENE_CLUSTER_E_VALUE_CUTOFF ){ //the cluster found has good evalue
					//insert a new cluster into the resulting cluster
					insert_cluster(p, p2, evals, i+1, n, clusters);					
				}

				p2 = p2->next;
				p  = p->next;
			} 
			else 
			{
				p = p->next;
				while(p != p2  && p2->pos - p->pos > i)
					p = p->next;

				if(p != p2  && p2->pos - p->pos == i){ //meet a cluster with window_size=i+1
					curr = p;
					n = 1;

					while(curr != p2){
						++n;
						curr = curr->next;
					}

					if( evals[i+1][n] <= GENE_CLUSTER_E_VALUE_CUTOFF ){ //the cluster found has good evalue
						//insert a new cluster into the resulting cluster
						insert_cluster(p, p2, evals, i+1, n, clusters);					
					}

					p2 = p2->next;
					p  = p->next;
				}
			}
		}
	}

	return;
}


//find gene cluster in a circular chromosome and get a pointer to T_RESULT containing resulting clusters
void find_clusters_inOneCircularChr(T_POSITION * pos, double ** evals, int maxwinsize, T_GCLUSTER ** clusters, int chrSize){
	int  i, n;

	T_POSITION * curr, *p, *p2; 

	T_POSITION * head, * tail;

	int passOrigin = 0; 

	head = pos; 

	tail = pos;
	while(tail->next)
	{
		tail = tail->next;
	}

	//now connect tail to head and form a circular list
	tail->next = head; 

	for(i=1; i<maxwinsize; ++i){ //window size = i+1
		p = pos;
		p2 = pos;

		passOrigin = 0;

		while( p2 ){
			int dist=0;

			while(p2){
				dist = p2->pos - p->pos;

				if( passOrigin ){
					dist += chrSize; 
				}

				if(dist < i){
					p2 = p2->next;

					if(p2 == head)
						passOrigin = 1;
				}
				else
					break;				
			}


			if( dist == i ){ //meet a cluster with window_size=i+1
				curr = p;
				n = 1;

				while(curr != p2){
					++n;
					curr = curr->next;
				}

				if( evals[i+1][n] <= GENE_CLUSTER_E_VALUE_CUTOFF ){ //the cluster found has good evalue
					//insert a new cluster into the resulting cluster
					insert_cluster(p, p2, evals, i+1, n, clusters);					
				}

				p2 = p2->next;
				if(p2 == head)
					passOrigin = 1;

				p  = p->next;
				if(p == head)
					break;
			} 
			else 
			{
				p = p->next;
				if(p == head){
					break;
				}

				while(p != p2){
					dist = p2->pos - p->pos;

					if( passOrigin ){
						dist += chrSize; 
					}

					if(dist > i){
						p = p->next;

						if(p == head)
							break;
					}
					else
						break;				
				}

				if(p == head)
					break;

				//while(p != p2  && p2->pos - p->pos > i){
				//	p = p->next;
				//}

				if(p != p2  && dist == i){ //meet a cluster with window_size=i+1
					curr = p;
					n = 1;

					while(curr != p2){
						++n;
						curr = curr->next;
					}

					if( evals[i+1][n] <= GENE_CLUSTER_E_VALUE_CUTOFF ){ //the cluster found has good evalue
						//insert a new cluster into the resulting cluster
						insert_cluster(p, p2, evals, i+1, n, clusters);					
					}

					p2 = p2->next;
					if(p2 == head)
						passOrigin = 1;

					p  = p->next;
					if(p == head)
						break;
				}
			}
		}
	}

	//now disconnect tail from head and go back to a linear list
	tail->next = NULL;

	return;
}


//remove repeats and subsets in resulting clusters
void remove_subclusters_inonechr(T_GCLUSTER * clusters){
	T_GCLUSTER *burr, *curr, *durr, *turr;

	 //remove repeats and subsets
	if( clusters != NULL && clusters->next != NULL ){
		 burr = clusters;
		 curr = clusters->next;
		 while( curr != NULL ){			 
			 durr = clusters;
			 while( durr != curr ){
				 if(durr->ps[0]->pos <= durr->ps[durr->n -1]->pos)
				 {
					 if( (curr->ps[0]->pos <= curr->ps[curr->n -1]->pos) && 
						 (durr->ps[0]->pos <= curr->ps[0]->pos) && 
						 (durr->ps[durr->n -1]->pos >= curr->ps[curr->n -1]->pos) )
					 {
						turr = curr;
						burr->next = curr->next;
						curr = curr->next;

						//free space
						if( turr->ps )
							free( turr->ps );
						if( turr )
							free( turr );

						break;
					}			
				 }
				 else
				 {
					 if( (durr->ps[0]->pos <= curr->ps[0]->pos && durr->ps[durr->n -1]->pos >= curr->ps[curr->n -1]->pos) || 
						 (durr->ps[0]->pos <= curr->ps[0]->pos && durr->ps[0]->pos <= curr->ps[curr->n -1]->pos) ||  
						 (curr->ps[0]->pos <= durr->ps[durr->n-1]->pos && curr->ps[curr->n-1]->pos <= durr->ps[durr->n-1]->pos))
					 {
						turr = curr;
						burr->next = curr->next;
						curr = curr->next;

						//free space
						if( turr->ps )
							free( turr->ps );
						if( turr )
							free( turr );

						break;
					}		
				 }

				 durr = durr->next;
			 }
			 
			 if ( durr == curr ) {
				 burr=burr->next;
				 curr=curr->next;
			 }			 
		 }
	}

	return;
}

//output resulting clusters
void output_clusters(T_PROTEIN * gcps, int n_gcps, T_GCLUSTER ** clusters, int n, char * resultfilename, char **chrFileList){
	FILE  *outf;
	int i, j, k;

	double min;

	T_GCLUSTER ** heads; 

	if( !(heads = malloc(n * sizeof(T_GCLUSTER *) )) )
		fprintf(stderr, "\n\nout of memory for heads!\n\n"), fflush(NULL), exit(0);

	if( !(outf = fopen(resultfilename, "w")) )
		fprintf(stderr, "File %s can't be open for writing\n", resultfilename), fflush(NULL), exit(0);

	for(i=0; i<n; ++i)
		heads[i] = clusters[i];

	do {
		min = 1.0e30;
		k = -1;

		for(i=0; i<n; ++i){
			if( heads[i] && heads[i]->e < min ){
				min = heads[i]->e;
				k = i;
			}
		}

		if( k >= 0 ){
			fprintf(outf, "\nExpect = %10.4e, size = %1d, chromosome:%s", heads[k]->e, heads[k]->n, chrFileList[k]);
			for(i=0; i<heads[k]->n; ++i){
				fprintf(outf, "\n%s [%d, ", heads[k]->ps[i]->name, heads[k]->ps[i]->pos);

				for(j=0; j<n_gcps; ++j){
					if( heads[k]->ps[i]->score[j] <= E_VALUE_CUTOFF ){
						fprintf(outf, "(%s %10.4e) ", gcps[j].name, heads[k]->ps[i]->score[j]);
					}
				}

				fprintf(outf, "]; ");				
			}
			fprintf(outf, "\n\n");

			heads[k] = heads[k]->next;
		} 
	} while (k >= 0);

	fclose(outf);

	return;
}


